// -*- java -*-
/***** BEGIN LICENSE BLOCK *****
Version: MPL 1.1/GPL 2.0/LGPL 2.1

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Initial Developer of the Original Code is Shawn Betts.
Portions created by the Initial Developer are Copyright (C) 2004.
Portions created by Stefan Hajnoczi are Copyright (C) 2006.
All Rights Reserved.

Contributor(s): Stefan Hajnoczi.

Alternatively, the contents of this file may be used under the terms of
either the GNU General Public License Version 2 or later (the "GPL"), or
the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
in which case the provisions of the GPL or the LGPL are applicable instead
of those above. If you wish to allow use of your version of this file only
under the terms of either the GPL or the LGPL, and not to allow others to
use your version of this file under the terms of the MPL, indicate your
decision by deleting the provisions above and replace them with the notice
and other provisions required by the GPL or the LGPL. If you do not delete
the provisions above, a recipient may use your version of this file under
the terms of any one of the MPL, the GPL or the LGPL.
***** END LICENSE BLOCK *****/

const gNumberedLinks = {
	GENERATE_DELAY: 250, // in milliseconds

	getState: function(doc) {
		if (gBrowser == undefined) {
			return null;
		}

		// Not sure why this sometimes throws a type error inside tabbrowser.xml
		var idx;
		try {
			idx = gBrowser.getBrowserIndexForDocument(doc);
		} catch (e) {
			return null;
		}

		if (idx == -1) {
			return null;
		}

		var currentBrowser = gBrowser.getBrowserAtIndex(idx);
		if (currentBrowser.numberedLinks == undefined) {
			currentBrowser.numberedLinks = {
				isEnabled: true
			};
		}

		var nl = currentBrowser.numberedLinks;
		nl.generateTimeout = null;
		return nl;
	},
	
	onLoad: function(evt) {
		var doc = evt.originalTarget;
		doc.addEventListener('DOMNodeInserted', gNumberedLinks.onDOMNodeInserted, false);
		gNumberedLinks.updateGenerateTimeout(doc);
	},

	onDOMNodeInserted: function(evt) {
		// Do not recurse on numbered link or other text nodes
		if (evt.target.nodeName == '#text') {
			return;
		}

		gNumberedLinks.updateGenerateTimeout(evt.currentTarget);
	},

	updateGenerateTimeout: function(doc) {
		var nl = gNumberedLinks.getState(doc);
		if (nl == null) {
			return;
		}

		if (nl.generateTimeout != null) {
			clearTimeout(nl.generateTimeout);
		}

		nl.generateTimeout = setTimeout(gNumberedLinks.generateLinks,
				gNumberedLinks.GENERATE_DELAY, doc);
	},

	generateLinks: function(doc) {
		var nl = gNumberedLinks.getState(doc);
		if (nl == null) {
			return;
		}

		nl.generateTimeout = null;

		if (nl.isEnabled) {
			// Remove and regenerate numbered links
			gNumberedLinks.setEnabled(doc, false);
			gNumberedLinks.setEnabled(doc, true);
		}
	},

	openBar: function() {
		var bar = document.getElementById('numberedlink-toolbar');
		var findfield = document.getElementById('numberedlink-field');
		bar.hidden = false;
		findfield.select();
		findfield.focus();
	},

	closeBar: function() {
		var findField = document.getElementById('numberedlink-field');
		var ww = Components.classes['@mozilla.org/embedcomp/window-watcher;1'].getService(Components.interfaces.nsIWindowWatcher);
		if (window == ww.activeWindow && document.commandDispatcher.focusedElement &&
			document.commandDispatcher.focusedElement.parentNode.parentNode == findField) {
			content.focus();
		}
		var bar = document.getElementById('numberedlink-toolbar');
		bar.hidden = true;
	},

	select: function(num) {
		gNumberedLinks.openBar();
		var field = document.getElementById('numberedlink-field');
		field.value = num;
	},

	// Send synthesized mouse event
	dispatchMouseEvent: function(doc, elem, type, evt) {
		var e = doc.createEvent('MouseEvents');
		e.initMouseEvent(type, true, true, doc.defaultView, 1, 0, 0, 0, 0,
				evt.ctrlKey, false, evt.shiftKey, evt.metaKey, 0, null);
		elem.dispatchEvent(e);
	},

	shouldClick: function(node) {
		if (node.tagName.match(/^a$/i)) {
			return true;
		}

		if (node.tagName.match(/^input$/i)) {
			return node.type.match(/^submit$/i) ||
					node.type.match(/^reset$/i) ||
					node.type.match(/^button$/i) ||
					node.type.match(/^checkbox$/i) ||
					node.type.match(/^radio$/i) ||
					node.type.match(/^image$/i);
		}

		return false;
	},

	shouldSelect: function(node) {
		if (node.tagName.match(/^input$/i)) {
			return node.type.match(/^checkbox$/i) ||
					node.type.match(/^radio$/i);
		}

		return false;
	},

	onKeyPress: function(evt) {
		if (evt.keyCode == KeyEvent.DOM_VK_RETURN) {
			var doc = window.content.document;
			var findfield = document.getElementById('numberedlink-field');
			var link = Number(findfield.value) - 1;
			var nodes = gNumberedLinks.getPageLinkNodes(doc);
			if (link >= 0 && link < nodes.length) {
				nodes[link].focus();
				if (!evt.altKey) {
					if (gNumberedLinks.shouldClick(nodes[link])) {
						gNumberedLinks.dispatchMouseEvent(doc, nodes[link], 'mouseover', evt);
						gNumberedLinks.dispatchMouseEvent(doc, nodes[link], 'mousedown', evt);
						gNumberedLinks.dispatchMouseEvent(doc, nodes[link], 'mouseup', evt);
						gNumberedLinks.dispatchMouseEvent(doc, nodes[link], 'click', evt);
					}
					if (gNumberedLinks.shouldSelect(nodes[link])) {
						nodes[link].select();
					}
				}
			}
			gNumberedLinks.closeBar();
		} else if (evt.keyCode == KeyEvent.DOM_VK_TAB ||
				evt.keyCode == KeyEvent.DOM_VK_ESCAPE) {
			gNumberedLinks.closeBar();
			evt.preventDefault();
		}
	},

	toggleNum: function(doc, node, number, enable) {
		var parent;
		var child;

		if (node.tagName.match(/^a$/i)) {
			parent = node;
			child = node.firstChild;
		} else {
			parent = node.parentNode;
			child = enable ? node : node.previousSibling;
		}

		if (enable) {
			var numText = doc.createTextNode('[' + number + ']');
			parent.insertBefore(numText, child);
		} else {
			// Make sure only to remove numbered links
			if (child && child.nodeName == '#text' && child.data.match(/^\[\d+\]$/)) {
				parent.removeChild(child);
			}
		}
	},

	getPageLinkNodes: function(doc) {
		var frames = doc.defaultView.frames;
		var nodes = gNumberedLinks.getLinkNodes(doc);
		for (var i = 0; i < frames.length; i++) {
			var tmp = gNumberedLinks.getLinkNodes(frames[i].document);
			for (var j = 0; j < tmp.length; j++) {
				nodes.push(tmp[j]);
			}
		}

		return nodes;
	},

	shouldLink: function(node) {
		if (node.tagName.match(/^a$/i)) {
			return node.hasAttribute('href');
		}

		if (node.tagName.match(/^input$/i)) {
			return !node.type.match(/^hidden$/i);
		}

		return true;
	},

	getLinkNodes: function(doc) {
		var tags = ['input', 'select', 'textarea', 'a'];
		var links = [];

		for (var i = 0; i < tags.length; i++) {
			var nodes = doc.getElementsByTagName(tags[i]);
			for (var j = 0; j < nodes.length; j++) {
				if (gNumberedLinks.shouldLink(nodes[j])) {
					links.push(nodes[j]);
				}
			}
		}

		return links;
	},

	setEnabled: function(doc, enable) {
		var nl = gNumberedLinks.getState(doc);

		// Keep track of the numbered state
		if (enable == nl.isEnabled) {
			return;
		}
		nl.isEnabled = enable;

		// Accumulate nodes and number them
		var nodes = gNumberedLinks.getPageLinkNodes(doc);
		for (var i = 0; i < nodes.length; i++) {
			gNumberedLinks.toggleNum(doc, nodes[i], i + 1, enable);
		}
	},

	toggle: function() {
		var doc = content.content.document;
		var nl = gNumberedLinks.getState(doc);
		gNumberedLinks.setEnabled(doc, !nl.isEnabled);
	}
};

window.addEventListener('load', gNumberedLinks.onLoad, true);
